# _abstract_archiver.py
#
# Project: AutoArchive
# License: GNU GPLv3
#
# Copyright (C) 2003 - 2012 Róbert Čerňanský



""":class:`_AbstractArchiver` class."""



__all__ = ["_AbstractArchiver"]



# {{{ INCLUDES

from abc import *

# }}}



# {{{ CLASSES

class _AbstractArchiver(metaclass = ABCMeta):
    """An archiver that creates :term:`backups <backup>` based on an *archive specification*\ .

    Abstract constructor of this class, should be called from derived constructors.

    :param interfaceAccessor: An :class:`.IInterfaceAccessor` instance.
    :type interfaceAccessor: :class:`.IInterfaceAccessor`"""

    #: Archiver type.  Type: :data:`.ArchiverTypes`.
    ARCHIVER_TYPE = None

    #: Suffix (extension) of files created by an archiver of type :data:`ARCHIVER_TYPE`.  Type: ``str``.
    SUFFIX_ = None



    @abstractmethod
    def __init__(self, interfaceAccessor):

        #: Stores :class:`.IComponentUi` instance.
        self.componentUi_ = None

        #: Stores :class:`._ArchiveSpec` instance.
        self.archiveSpec_ = None

        #: Persistent storage for an archive
        self.storagePortion_ = None

        # derived classes has to specify self.ARCHIVER_ and self.SUFFIX_
        assert self.ARCHIVER_TYPE is not None and self.SUFFIX_



    @abstractmethod
    def attachArchiveSpec(self, archiveSpec):
        """Attaches an :term:`archive` to the :term:`archiver`.

        This method has to be called prior to any operation with the archive.

        See also: :meth:`detachArchiveSpec()`.

        :param archiveSpec: The *archive specification* representing the archive that shall be attached.
        :type archiveSpec: :class:`._ArchiveSpec`

        :raise RuntimeError: If there is already some archive attached."""



    @abstractmethod
    def detachArchiveSpec(self):
        """Detaches an :term:`archive` from the :term:`archiver`.

        The :term:`archive` that was previously attached by the :meth:`attachArchiveSpec()` method will be detached,
        returning this instance to the state after its construction.

        See also: :meth:`attachArchiveSpec()`.

        :raise RuntimeError: If there is no archive attached."""

    

    @abstractmethod
    def createBackup(self):
        """Run the :term:`archiver` and create a :term:`backup` based on the attached archive specification."""



    @classmethod
    @abstractmethod
    def getStoredArchiveNames(cls, userConfigDir, storage):
        """Returns iterable of archive names which has some data stored in a persistent storage.

        Persistent storages from which archive names are retrieved are specific to the concrete archiver type.  A common
        storage used by archivers is the application storage (:class:`.IStorage`).

        :param userConfigDir: Path to the user configuration directory.
        :type userConfigDir: ``str``
        :param storage: The application storage.
        :type storage: :class:`.IStorage`

        :return: Iterable of archive names.
        :rtype: ``Iterable<str>``"""



    @classmethod
    @abstractmethod
    def tryPurgeStoredArchiveData(cls, archiveName, userConfigDir, storage):
        """Deletes all data stored for the archive named ``archiveName`` if any.

        See also: :meth:`getStoredArchiveNames()`

        :param archiveName: Name of the archive which data shall be purged.
        :type archiveName: ``str``
        :param userConfigDir: Path to the user configuration directory.
        :type userConfigDir: ``str``
        :param storage: The application storage.
        :type storage: :class:`.IStorage`

        :return: ``True`` if data was purged; ``False`` otherwise.
        :rtype: ``bool``

        :raise OSError: If an error occurred during the operation of removing data from a physical storage."""



    @abstractproperty
    def archiveFileName_(self):
        """Gets path to the file that will be created by the archiver.

        :rtype: ``str``"""

# }}} CLASSES
