# _archiver_base.py
#
# Project: AutoArchive
# License: GNU GPLv3
#
# Copyright (C) 2003 - 2012 Róbert Čerňanský



""":class:`_ArchiverBase` class."""



__all__ = ["_ArchiverBase"]



# {{{ INCLUDES

from abc import *

from ..._mainf import *
from ..._configuration import *
from ._archive_spec import *
from ._abstract_archiver import *

# }}} INCLUDES



# {{{ CLASSES

class _ArchiverBase(_AbstractArchiver):
    """Base class for archivers.

    Abstract constructor of this class, should be called from derived constructors.  It initializes the
    :attr:`._AbstractArchiver.componentUi_` attribute.

    :param interfaceAccessor: An :class:`.IInterfaceAccessor` instance.
    :type interfaceAccessor: :class:`.IInterfaceAccessor`"""

    #: Iterable of options not supported by the archive.
    UNSUPPORTED_OPTIONS_ = frozenset()



    # {{{ _AbstractArchiver implementation

    @abstractmethod
    def __init__(self, interfaceAccessor):
        super().__init__(interfaceAccessor)

        #: Stores :class:`.IComponentUi` instance.
        self.componentUi_ = interfaceAccessor.getComponentInterface(IComponentUi)

        self.__storage = interfaceAccessor.getComponentInterface(IStorage)



    def attachArchiveSpec(self, archiveSpec):
        "See: :meth:`._AbstractArchiver.attachArchiveSpec()`"

        if self.archiveSpec_ is not None:
            raise RuntimeError("There is already an archive specification file attached.")

        self.archiveSpec_ = archiveSpec
        self.storagePortion_ = self.__storage.createStoragePortion(realm = self.archiveSpec_[_ArchiveSpecOptions.NAME])



    def detachArchiveSpec(self):
        "See: :meth:`._AbstractArchiver.detachArchiveSpec()`"

        if self.archiveSpec_ is None:
            raise RuntimeError("No archive specification file attached.")

        self.storagePortion_ = None
        self.archiveSpec_ = None



    def createBackup(self):
        """See: :meth:`._AbstractArchiver.createBackup()`.

        Method calls :meth:`preExecution_()` followed by :meth:`executeArchiver_()` and :meth:`postExecution_()`."""

        self.__reportUnsupportedOptions()
        self.preExecution_()
        success = self.executeArchiver_()
        self.postExecution_(success)



    @classmethod
    def getStoredArchiveNames(cls, userConfigDir, storage):
        "See: :meth:`._AbstractArchiver.getStoredArchiveNames()`"

        return storage.getRealms()



    @classmethod
    def tryPurgeStoredArchiveData(cls, archiveName, userConfigDir, storage):
        "See: :meth:`._AbstractArchiver.tryPurgeStoredArchiveData()`"

        try:
            storage.removeRealm(archiveName)
            return True
        except KeyError:
            return False

    # }}} _AbstractArchiver implementation



    # {{{ virtual members

    @property
    def suffix_(self):
        """Gets the backup file name suffix.

        :rtype: ``str``"""

        return self.SUFFIX_



    @abstractmethod
    def executeArchiver_(self):
        """Creates the backup.

        :return: ``False`` if creation of the backup failed.
        :rtype: ``bool``"""



    def preExecution_(self):
        """Do whatever is necessary before :term:`archiver` execution.

        See also: :meth:`createBackup()`."""

        pass



    def postExecution_(self, success):
        """Do whatever is necessary after :term:`archiver` execution.

        See also: :meth:`createBackup()`.

        :param success: ``True``, if the backup was created successfully.
        :type success: ``bool``"""

        pass



    def handleArchiverExitCode_(self, exitCode):
        """Handles archiver's exit code.

        .. warning:: This method has to be called only to handle non-zero exit code.
        .. note:: In case of error or warning, this method should call the appropriate :class:`.IComponentUi`\
           .\ `show*()` method because UI may rely on it to calculate overall status of the archiving operation.

        :param exitCode: Archiver's non-zero exit code.
        :type exitCode: ``int``

        :return: ``False``, if ``exitCode`` means that the backup creation failed; ``True`` otherwise.
        :rtype: ``bool``"""

        self.componentUi_.showError("Creation of the backup failed!")
        return False

    # }}} virtual members



    # {{{ helpers

    def __reportUnsupportedOptions(self):
        for option in self.UNSUPPORTED_OPTIONS_:
            if self.archiveSpec_.getRawValue(option) is not None:
                self.componentUi_.showWarning(str.format(
                    "Option \"{}\" is not supported by the archiver of type \"{}\".",
                    option, OptionsUtils.archiverTypeToStr(self.ARCHIVER_TYPE)))

    # }}} helpers

# }}} CLASSES
