# _external_archivers.py
#
# Project: AutoArchive
# License: GNU GPLv3
#
# Copyright (C) 2003 - 2012 Róbert Čerňanský



"""Concrete classes derived from :class:`._ExternalArchiverBase`."""



__all__ = ["_ExternalTarArchiver", "_ExternalTarCompressedArchiver",
           "_ExternalTarGzArchiver", "_ExternalTarBz2Archiver",
           "_ExternalTarLzmaArchiver", "_ExternalTarXzArchiver"]



# {{{ INCLUDES

from abc import *
import os
import re
import glob
import shutil
import itertools

from ..._utils import *
from ..._configuration import *
from .. import *
from ._archive_spec import *
from ._external_archiver_base import *
from ._incremental_archiver_mixin import *

# }}} INCLUDES


# {{{ CLASSES

class _ExternalTarArchiver(_ExternalArchiverBase, _IncrementalArchiverMixin):
    "Represents external GNU tar archiver."

    #: See: :attr:`._AbstractArchiver.ARCHIVER_TYPE`.
    ARCHIVER_TYPE = ArchiverTypes.Tar

    #: See: :attr:`._AbstractArchiver.SUFFIX_`.
    SUFFIX_ = ".tar"

    #: See: :attr:`._ArchiverBase.UNSUPPORTED_OPTIONS_`.
    UNSUPPORTED_OPTIONS_ = frozenset({Options.COMPRESSION_LEVEL})

    #: See: :attr:`._ExternalArchiverBase.ARCHIVER_`.
    ARCHIVER_ = "/bin/tar"

    # subdirectory for snapshots
    __SNAPSHOTS_SUBDIR = "snapshots"

    # suffix for snapshot files used in incremental backups
    __SNAPSHOT_SUFFIX = ".snar"



    # {{{ overrides of _ArchiverBase

    def __init__(self, interfaceAccessor):
        super().__init__(interfaceAccessor)

        appConfig = interfaceAccessor.getComponentInterface(IAppConfig)
        self.__snapshotsDir = self.__getSnapshotsDir(appConfig[Options.USER_CONFIG_DIR])
        self.tryMakeSnapshotsDir(appConfig[Options.USER_CONFIG_DIR], self.componentUi_)



    def attachArchiveSpec(self, archiveSpec):
        "See: :meth:`._ArchiverBase.attachArchiveSpec()`."

        super().attachArchiveSpec(archiveSpec)
        self.setRestartLevel_()
        self.setNextBackupLevel_()



    def detachArchiveSpec(self):
        "See: :meth:`._ArchiverBase.detachArchiveSpec()`."

        self.clearNextBackupLevel_()
        self.clearRestartLevel_()
        super().detachArchiveSpec()



    @classmethod
    @_Utils.uniq
    def getStoredArchiveNames(cls, userConfigDir, storage):
        "See: :meth:`._ArchiverBase.getStoredArchiveNames()`."

        snapshots = cls.__getSnapshots(cls.__getSnapshotsDir(userConfigDir))
        names = (os.path.splitext(os.path.splitext(os.path.basename(snapshot))[0])[0]
                 for snapshot in snapshots)
        return itertools.chain(super().getStoredArchiveNames(userConfigDir, storage), names)



    @classmethod
    def tryPurgeStoredArchiveData(cls, archiveName, userConfigDir, storage):
        "See: :meth:`._ArchiverBase.tryPurgeStoredArchiveData()`."

        snapshotsDir = cls.__getSnapshotsDir(userConfigDir)
        snapshots = cls.__getSnapshots(snapshotsDir, archiveName)
        for snapshot in snapshots:
            os.remove(os.path.join(snapshotsDir, snapshot))
        return super().tryPurgeStoredArchiveData(archiveName, userConfigDir, storage) or len(snapshots) > 0



    @property
    def archiveFileName_(self):
        "See: :attr:`._AbstractArchiver.archiveFileName_`."

        return self.__getBackupFileName(self.nextBackupLevel)



    def preExecution_(self):
        "See: :meth:`._ArchiverBase.preExecution_()`."

        self.__manageSnapshotFiles()
        self.reportBackupLevelRestart_()
        super().preExecution_()



    def postExecution_(self, success):
        "See: :meth:`._ArchiverBase.postExecution_()`."

        if success:
            self.saveRestartingInfo_()
        super().postExecution_(success)

    # }}} overrides of _ArchiverBase

    

    # {{{ overrides of _ExternalArchiverBase

    def arguments_(self, sysEnvironment):
        "See: :meth:`._ExternalArchiverBase.arguments_()`."

        # operation has to be first one
        archiverOptions = ["--create", "--format=posix"]

        # add converted options passed to aa
        archiverOptions += self.__convertOptions()

        # add options required for this archiver type
        archiverOptions += ["--file=" + self.archiveFileName_,
                            "--directory=" + self.archiveSpec_[_ArchiveSpecOptions.PATH]]

        # add converted include and exclude files
        archiverOptions += self.__convertArchiveSpec()

        return archiverOptions, {}



    def showArchiverMessage_(self, message, sentToStderr = False):
        """Shows archiver message as an error, warning, notification, etc.

        Parses the passed ``message``, evaluates it and shows it as the message of appropriate verbosity level/type.
        If the message does not match any of supported patterns then default handling of the super class is used.

        See also: :meth:`._ExternalArchiverBase.showArchiverMessage_()`."""

        if sentToStderr and re.search("cannot (open|stat):", message, re.IGNORECASE):
            self.componentUi_.showError(message)
        else:
            super().showArchiverMessage_(message, sentToStderr)



    def handleArchiverExitCode_(self, exitCode):
        "See: :meth:`._ArchiverBase.handleArchiverExitCode_()`."

        if exitCode == 1:
            self.componentUi_.showWarning("Some files were changed during the backup creation.")
            return True

        return super().handleArchiverExitCode_(exitCode)

    # }}} overrides of _ExternalArchiverBase



    # {{{ overrides of _IncrementalArchiverMixin

    @classmethod
    def getBackupLevelForArchive(cls, archiveName, userConfigDir):
        "See: :meth:`._IncrementalArchiverMixin.getBackupLevelForArchive()`."

        level = None
        snapshotFiles = cls.__getSnapshots(cls.__getSnapshotsDir(userConfigDir), archiveName)

        if len(snapshotFiles) > 0:
            level = max((cls.getLevelFromFileName_(snapshotFile)
                         for snapshotFile in snapshotFiles))

        return level

    # }}} overrides of _IncrementalArchiverMixin



    @classmethod
    def tryMakeSnapshotsDir(cls, userConfigDir, componentUi):
        """Creates the snapshots directory.  Shows error if creation fails.

        .. note:: This method should be called once, before calling either of :meth:`getStoredArchiveNames()` or \
           :meth:`getBackupLevelForArchive()`\ methods.  It is not necessary to call it if constructor of this class has
           been already called.

        :param userConfigDir: Path to the user configuration directory.
        :type userConfigDir: ``str``
        :param componentUi: Component UI instance.
        :type componentUi: :class:`.IComponentUi`"""

        snapshotsDir = cls.__getSnapshotsDir(userConfigDir)
        if not os.path.exists(snapshotsDir):
            try:
                os.mkdir(snapshotsDir)
            except OSError as ex:
                componentUi.showError(str.format(
                        "Unable to create snapshots subdirectory. The reason is: {}", ex.strerror))



    # {{{ helpers

    # {{{ argument generation helpers

    def __convertOptions(self):
        """Returns list of options for archiver.

        Converts options passed to the program to similar options for the archiver program."""

        archiverOptions = []

        if self.archiveSpec_[Options.VERBOSE]:
            archiverOptions.append("--verbose")

        if self.archiveSpec_[Options.INCREMENTAL]:
            archiverOptions.append("--listed-incremental=" + self.__getSnapshotFileName(self.nextBackupLevel))

        return archiverOptions



    def __convertArchiveSpec(self):
        """Returns list of arguments for the archiver.

        Converts list of files and list of excluded files to the form suitable for the archiver program."""

        archiverOptions = []

        if self.archiveSpec_[_ArchiveSpecOptions.EXCLUDE_FILES]:
            archiverOptions.append("--anchored")
        for exclude in self.archiveSpec_[_ArchiveSpecOptions.EXCLUDE_FILES]:
            archiverOptions.append("--exclude=" + exclude)

        archiverOptions += self.archiveSpec_[_ArchiveSpecOptions.INCLUDE_FILES]

        return archiverOptions

    # }}} argument generation helpers



    # {{{ incremental backup helpers

    @classmethod
    def __getSnapshots(cls, snapshotsDir, archiveName = ""):
        """Returns sequence of snapshot file names for the archive named ``archiveName`` or all of them.

        :param snapshotsDir: Directory where the snapshot files are stored.  Can be obtained with
           :meth:`_getSnapshotsDir` method.
        :type snapshotsDir: ``str``
        :param archiveName: Name of the archive for which the snapshot file names shall be returned.  If not specified
           all snapshot files will be returned.
        :type archiveName: ``str``

        :return: Sequence of snapshot file names.
        :rtype: ``Sequence<str>``"""

        if archiveName == "":
            archiveName = "*"
        snapshots = itertools.chain(
            glob.iglob(os.path.join(snapshotsDir, archiveName + cls.__SNAPSHOT_SUFFIX)),
            glob.iglob(os.path.join(snapshotsDir, archiveName + ".*" + cls.__SNAPSHOT_SUFFIX)))
        return tuple(os.path.basename(snapshot)
                     for snapshot in snapshots)



    def __getSnapshotsForCurrentArchive(self):
        "Returns sequence of snapshot file names for the archive that this instance is processing."

        return self.__getSnapshots(self.__snapshotsDir, self.archiveSpec_[_ArchiveSpecOptions.NAME])



    @classmethod
    def __getSnapshotsDir(cls, userConfigDir):
        return os.path.join(userConfigDir, cls.__SNAPSHOTS_SUBDIR)



    def __manageSnapshotFiles(self):
        if not self.archiveSpec_[Options.INCREMENTAL]:
            return

        if self.nextBackupLevel > 0:

            # copy the old snapshot file to a new one so it will be possible to create also lower level backups
            # (otherwise it would be only possible to create level N+1 backup (where N is the latest/current level) or
            # level 0 (go from the beginning))
            shutil.copyfile(self.__getSnapshotFileName(self.nextBackupLevel - 1),
                            self.__getSnapshotFileName(self.nextBackupLevel))

        else:

            # we are going to create level 0 backup so remove snapshot for level 0 because new one will be created
            if os.path.exists(self.__getSnapshotFileName(0)):
                os.remove(self.__getSnapshotFileName(0))

        # remove higher level snapshot files
        for snapshot in self.__getSnapshotsForCurrentArchive():
            if self.getLevelFromFileName_(snapshot) > self.nextBackupLevel:
                os.remove(os.path.join(self.__snapshotsDir, snapshot))



    def __getBackupFileName(self, level):
        """Gets path to the file that shall be created by the :term:`archiver` program.

        :param level: Backup level for which the file name shall be returned.
        :type level: ``int``

        :return: Path to the backup file.
        :rtype: ``str``"""

        destDir = self.archiveSpec_[Options.DEST_DIR]
        fileName = self.archiveSpec_[_ArchiveSpecOptions.NAME] + self.getLevelSuffix_(level) + self.suffix_

        return os.path.join(destDir, fileName)



    def __getSnapshotFileName(self, level):
        "Returns full path to snapshot file for a certain backup level."

        return os.path.join(
            self.__snapshotsDir,
            self.archiveSpec_[_ArchiveSpecOptions.NAME] +
            self.getLevelSuffix_(level) +
            self.__SNAPSHOT_SUFFIX)

    # }}} incremental backup helpers

    # }}} helpers



class _ExternalTarCompressedArchiver(_ExternalTarArchiver):
    "Represents GNU tar archiver with a compression."

    #: See: :attr:`._ArchiverBase.UNSUPPORTED_OPTIONS_`.
    UNSUPPORTED_OPTIONS_ = frozenset()

    #: Suffix specific for a compressed tar backup file.
    COMPRESS_SUFFIX_ = None

    #: Tar option for creating specific compressed backup.
    COMPRESS_OPTION_ = None

    #: Environment variable from which the compress program reads its options.
    ENV_NAME_ = None



    # {{{ overrides of _ExternalTarArchiver

    @abstractmethod
    def __init__(self, interfaceAccessor):
        super().__init__(interfaceAccessor)

        # derived classes has to specify fields below
        assert self.COMPRESS_SUFFIX_ and self.COMPRESS_OPTION_



    def arguments_(self, sysEnvironment):
        "See: :meth:`._ExternalTarArchiver.arguments_`."

        # get arguments from base class
        baseOptions, baseEnvironment = super().arguments_(sysEnvironment)

        # convert options passed to aa to arguments
        archiverOptions, environment = self.__convertOptions(sysEnvironment)

        # insert options required for this archiver
        archiverOptions.append(self.COMPRESS_OPTION_)

        # assemble options so first item from baseOptions remains first
        # because it is the operation
        archiverOptions = baseOptions[:1] + archiverOptions + baseOptions[1:]

        # add environment from base class
        environment.update(baseEnvironment)

        return archiverOptions, environment

    # }}} overrides of _ExternalTarArchiver



    # {{{ overrides of _ArchiverBase

    @property
    def suffix_(self):
        "See: :obj:`._ArchiverBase.suffix_`."

        return super().suffix_ + self.COMPRESS_SUFFIX_

    # }}} overrides of _ArchiverBase



    # {{{ helpers

    def __convertOptions(self, sysEnvironment):
        """Returns tuple of options and environment for archiver.

        Converts options passed to aa to similar options and environment for the archiver program."""

        environment = {}

        if (self.archiveSpec_[Options.COMPRESSION_LEVEL] is not None) and self.ENV_NAME_:
            sysEnvValue = ""
            if self.ENV_NAME_ in sysEnvironment:
                sysEnvValue = sysEnvironment[self.ENV_NAME_] + " "

            environment[self.ENV_NAME_] = sysEnvValue + "-" + str(self.archiveSpec_[Options.COMPRESSION_LEVEL])

        return [], environment

    # }}} helpers



class _ExternalTarGzArchiver(_ExternalTarCompressedArchiver):
    "Represents GNU tar archiver with gzip compression."

    #: See: :attr:`._AbstractArchiver.ARCHIVER_TYPE`.
    ARCHIVER_TYPE = ArchiverTypes.TarGz

    #: See: :attr:`_ExternalTarCompressedArchiver.COMPRESS_SUFFIX_`.
    COMPRESS_SUFFIX_ = ".gz"

    #: See: :attr:`_ExternalTarCompressedArchiver.COMPRESS_OPTION_`.
    COMPRESS_OPTION_ = "--gzip"

    #: See: :attr:`_ExternalTarCompressedArchiver.ENV_NAME_`.
    ENV_NAME_ = "GZIP"



    def __init__(self, interfaceAccessor):
        super().__init__(interfaceAccessor)



class _ExternalTarBz2Archiver(_ExternalTarCompressedArchiver):
    "Represents GNU tar archiver with bzip2 compression."

    #: See: :attr:`._AbstractArchiver.ARCHIVER_TYPE`.
    ARCHIVER_TYPE = ArchiverTypes.TarBz2

    #: See: :attr:`_ExternalTarCompressedArchiver.COMPRESS_SUFFIX_`.
    COMPRESS_SUFFIX_ = ".bz2"

    #: See: :attr:`_ExternalTarCompressedArchiver.COMPRESS_OPTION_`.
    COMPRESS_OPTION_ = "--bzip2"

    #: See: :attr:`_ExternalTarCompressedArchiver.ENV_NAME_`.
    ENV_NAME_ = "BZIP2"



    def __init__(self, interfaceAccessor):
        super().__init__(interfaceAccessor)



class _ExternalTarLzmaArchiver(_ExternalTarCompressedArchiver):
    "Represents GNU tar archiver with lzma compression."

    #: See: :attr:`._AbstractArchiver.ARCHIVER_TYPE`.
    ARCHIVER_TYPE = ArchiverTypes.TarLzma

    #: See: :attr:`_ExternalTarCompressedArchiver.COMPRESS_SUFFIX_`.
    COMPRESS_SUFFIX_ = ".lzma"

    #: See: :attr:`_ExternalTarCompressedArchiver.COMPRESS_OPTION_`.
    COMPRESS_OPTION_ = "--lzma"



    def __init__(self, interfaceAccessor):
        super().__init__(interfaceAccessor)



class _ExternalTarXzArchiver(_ExternalTarCompressedArchiver):
    "Represents GNU tar archiver with xz compression."

    #: See: :attr:`._AbstractArchiver.ARCHIVER_TYPE`.
    ARCHIVER_TYPE = ArchiverTypes.TarXz

    #: See: :attr:`_ExternalTarCompressedArchiver.COMPRESS_SUFFIX_`.
    COMPRESS_SUFFIX_ = ".xz"

    #: See: :attr:`_ExternalTarCompressedArchiver.COMPRESS_OPTION_`.
    COMPRESS_OPTION_ = "--use-compress-program=xz"

    #: See: :attr:`_ExternalTarCompressedArchiver.ENV_NAME_`.
    ENV_NAME_ = "XZ_OPT"



    def __init__(self, interfaceAccessor):
        super().__init__(interfaceAccessor)

# }}} CLASSES
