# icomponent_ui.py
#
# Project: AutoArchive
# License: GNU GPLv3
#
# Copyright (C) 2003 - 2011 Róbert Čerňanský



""":class:`IComponentUi` interface and :data:`UiMessageKinds` enum."""



__all__ = ["UiMessageKinds", "IComponentUi"]



# {{{ INCLUDES

from abc import *

from .._py_additions import *
from .iinterface_accessor import *

# }}} INCLUDES



# {{{ CONSTANTS

#: Kinds of user messages.
UiMessageKinds = Enum(
    "Verbose",
    "Notification",
    "Info",
    "Warning",
    "Error")

# }}} CONSTANTS



# {{{ CLASSES

class IComponentUi(IComponentInterface):
    """Basic interface for an :term:`UI` implementation.

    Defines basic methods for showing messages of various importance to a user.

    .. note:: There should be at least one :term:`component` that implements this interface in order to provide an \
       :term:`UI` access to other components.

    See also the description of _mainf package (:mod:`._mainf`)."""

    @abstractmethod
    @event
    def messageShown(messageKind):
        """Fired after a user message was shown.

        :param messageKind: Kind of the message that was shown.
        :type messageKind: :data:`UiMessageKinds`"""



    @abstractproperty
    def verbosity(self):
        """Gets the verbosity level.

        :rtype: ``int``"""



    @abstractmethod
    def showVerbose(self, msg):
        """Show a verbose-type message to a user.

        Verbose messages should be shown only if user enables it.  Although this method can be called regardless of
        current verbosity level, the concrete implementation can decide whether it will be shown or not if verbosity
        level is 0.  It is not recommended to call this method if verbosity level is 0 due to performance reasons.
        Current verbosity level can be obtained via :attr:`verbosity` property.

        See also: :data:`verbosity`.

        :param msg: The message that should be shown to the user.
        :type msg: ``str``"""



    @abstractmethod
    def showNotification(self, msg):
        """Show an unintrusive notification message to a user.

        .. note:: If user interface implementation does not have means to support notifications then it should be
           presented to the user similarly as :meth:`showInfo`.

        :param msg: The message that should be shown to the user.
        :type msg: ``str``"""



    @abstractmethod
    def showInfo(self, msg):
        """Show an information message to a user.

        :param msg: The message that should be shown to the user.
        :type msg: ``str``"""



    @abstractmethod
    def showWarning(self, msg):
        """Show a warning message to a user.

        :param msg: The message that should be shown to the user.
        :type msg: ``str``"""



    @abstractmethod
    def showError(self, msg):
        """Show an error message to a user.

        :param msg: The message that should be shown to the user.
        :type msg: ``str``"""

# }}} CLASSES
